/* $Id: txm.c,v 1.17 2006/08/07 15:55:51 geni Exp $
 *
 * Copyright (c) 2005 Huidae Cho
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "onhand.h"
#include "define.h"

static void txm_rec_qsort(txm_rec_t *head, txm_rec_t *tail);

int
txm_read(txm_t *db, char *file)
{
	int size, i;
	u8 *data, *ptr;
	u32 offset;
	txm_rec_t *rec;

	if(read_file(file, &data, &size)){
		ERROR("read_file");
		return -1;
	}
	/* the minimum size is 16 = 4+2+2+2*4 */
	if(size < 16 || memcmp(data, pim_magic, 4)){
		ERROR("%s: not an txm file", file);
		return -1;
	}

	txm_init(db);

	memcpy(&offset, data+8, 4);
	while(offset != 0xffffffff && offset < size){
		db->num_recs++;
		ptr = data + offset;
		memcpy(&offset, ptr+4, 4);
		rec = (txm_rec_t *)zalloc(sizeof(txm_rec_t));
		rec->priv = ptr[14];
		if(rec->priv)
			db->num_priv++;
		ptr += 18;
		for(i = 0; i < TXM_NUM_FIELDS; i++){
			memcpy(&rec->data[i].size, ptr, 2);
			ptr += 2;
			rec->data[i].buf = NULL;
			if(!rec->data[i].size)
				continue;
			rec->data[i].buf = (u8 *)malloc(rec->data[i].size+1);
			memcpy(rec->data[i].buf, ptr, rec->data[i].size);
			rec->data[i].buf[rec->data[i].size] = 0;
			ptr += rec->data[i].size;
		}
		rec->prev = rec->next = NULL;
		if(db->head){
			rec->prev = db->tail;
			db->tail->next = rec;
			db->tail = rec;
		}else
			db->head = db->tail = rec;
	}

	free(data);

	txm_rec_sort(db);

	return 0;
}

int
txm_write(txm_t *db, char *file)
{
	int fd, i;
	u8 i8;
	u16 i16, size;
	u32 i32, offset, offset2;
	txm_rec_t *rec;

	for(rec = db->head; rec && rec->data[TXM_TITLE].buf; rec = rec->next);
	if(rec && !rec->data[TXM_TITLE].buf){
		ERROR("internal error");
		return -1;
	}

	_open();

	_write(pim_magic, 4);
	_write(&db->num_priv, 2);
	_write(&db->num_recs, 2);

	/* header */
	if(db->head){
		offset = 16;
		_write(&offset, 4);
		for(rec = db->head; rec; rec = rec->next){
			offset2 = offset;
			offset += 18;
			for(i = 0; i < TXM_NUM_FIELDS; i++)
				offset += 2 + rec->data[i].size;
		}
		_write(&offset2, 4);
	}else{
		offset2 = 0xffffffff;
		for(i = 0; i < 2; i++)
			_write(&offset2, 4);
	}

	/* records */
	offset = 16;
	offset2 = 0xffffffff;
	for(rec = db->head; rec; rec = rec->next){
		_write(&offset2, 4);
		offset2 = offset;
		offset += 18;
		size = 0;
		for(i = 0; i < TXM_NUM_FIELDS; i++)
			size += 2 + rec->data[i].size;
		offset += size;
		if(rec->next)
			_write(&offset, 4);
		else{
			offset2 = 0xffffffff;
			_write(&offset2, 4);
		}
		i32 = 0;
		_write(&i32, 4);

		/* TODO: 00 ... 00 for unknown does not work! */
		i16 = 0xcb6e;
		_write(&i16, 2);

		_write(&rec->priv, 1);
		i8 = 0;
		_write(&i8, 1);
		_write(&size, 2);
		for(i = 0; i < TXM_NUM_FIELDS; i++){
			_write(&rec->data[i].size, 2);
			if(!rec->data[i].size)
				continue;
			_write(rec->data[i].buf, rec->data[i].size);
		}
	}

	_close();

	return 0;
}

void
txm_scan(txm_t *db, FILE *fp)
{
	int i, j;
	char *line;
	txm_rec_t *rec;

	rec = NULL;
	while((line = read_line(fp))){
		if(!line[0]){
			if(rec){
				if(rec->data[TXM_TITLE].buf){
					txm_rec_add(db, rec);
					if(rec->priv)
						db->num_priv++;
					db->num_recs++;
				}
				rec = NULL;
			}
		}else
		if(!strcmp(line, "private: yes")){
			if(!rec)
				rec = (txm_rec_t *)zalloc(sizeof(txm_rec_t));
			rec->priv = 1;
		}else
		for(i = 0; i < TXM_NUM_FIELDS; i++){
			j = strlen(txm_field[i]);
			if(strstr(line, txm_field[i]) != line ||
					strstr(line, ": ") != line + j)
				continue;
			if(!rec)
				rec = (txm_rec_t *)zalloc(sizeof(txm_rec_t));
			rec->data[i].size = strlen(line + j + 2);
			if(rec->data[i].buf)
				free(rec->data[i].buf);
			rec->data[i].buf = (u8 *)malloc(rec->data[i].size + 1);
			strcpy(rec->data[i].buf, line + j + 2);
			sub_chr(rec->data[i].buf, '\t', '\n');
		}
		free(line);
	}

	return;
}

void
txm_print(txm_t *db, FILE *fp)
{
	int i;
	txm_rec_t *rec;

	for(rec = db->head; rec; rec = rec->next){
		for(i = 0; i < TXM_NUM_FIELDS; i++){
			if(rec->data[i].buf){
				fprintf(fp, "%s: ", txm_field[i]);
				print_str(fp, rec->data[i].buf);
				fprintf(fp, "\n");
			}
		}
		if(rec->priv)
			fprintf(fp, "private: yes\n");
		fprintf(fp, "\n");
	}

	return;
}

void
txm_init(txm_t *db)
{
	db->num_priv = db->num_recs = 0;
	db->head = db->tail = NULL;

	return;
}

void
txm_free(txm_t *db)
{
	txm_rec_t *rec, *rec_next;

	for(rec = db->head; rec; rec = rec_next){
		rec_next = rec->next;
		txm_rec_free(rec);
	}

	return;
}

void
txm_rec_free(txm_rec_t *rec)
{
	int i;

	for(i = 0; i < TXM_NUM_FIELDS; i++){
		if(rec->data[i].buf)
			free(rec->data[i].buf);
	}
	free(rec);

	return;
}

DEFINE_ADD(txm, txm_rec, )

DEFINE_DEL(txm, txm_rec, )

DEFINE_QSORT(txm_rec)

void
txm_rec_sort(txm_t *db)
{
	txm_rec_qsort(db->head, db->tail);

	return;
}

int
txm_rec_cmp(txm_rec_t *a, txm_rec_t *b)
{
	return strcmp(a->data[TXM_TITLE].buf, b->data[TXM_TITLE].buf);
}
